@@ -4,8 +4,185 @@ use mod_perl qw(1.07 StackedHandlers MethodHandlers Authen Authz);
 use Apache::Constants qw(:common M_GET M_POST AUTH_REQUIRED REDIRECT);
 use vars qw($VERSION);
 
-$VERSION = substr(q$Revision: 1.4 $, 10);
- 
+# $Id: AuthCookie.pm,v 2.0 2000/02/02 13:18:23 ken Exp $
+$VERSION = sprintf '%d.%03d', q$Revision: 2.0 $ =~ /: (\d+).(\d+)/;
+
+sub recognize_user ($$) {
+  my ($self, $r) = @_;
+  my $debug = $r->dir_config("AuthCookieDebug") || 0;
+  my ($auth_type, $auth_name) = ($r->auth_type, $r->auth_name);
+  return unless $auth_type && $auth_name;
+  return unless $r->header_in('Cookie');
+
+  my ($cookie) = $r->header_in('Cookie') =~ /${auth_type}_${auth_name}=([^;]+)/;
+  $r->log_error("cookie ${auth_type}_${auth_name} is $cookie") if $debug >= 2;
+  if (my ($user) = $auth_type->authen_ses_key($r, $cookie)) {
+    $r->log_error("user is $user") if $debug >= 2;
+    $r->connection->user($user);
+  }
+}
+
+
+sub login ($$) {
+  my ($self, $r) = @_;
+  my $debug = $r->dir_config("AuthCookieDebug") || 0;
+
+  my ($auth_type, $auth_name) = ($r->auth_type, $r->auth_name);
+  my %args = $r->args;
+  unless (exists $args{'destination'}) {
+    $r->log_error("No key 'destination' found in posted data");
+    return SERVER_ERROR;
+  }
+  
+  # Get the credentials from the data posted by the client
+  my @credentials;
+  while (exists $args{"credential_" . ($#credentials + 1)}) {
+    $r->log_error("credential_" . ($#credentials + 1) . " " .
+		  $args{"credential_" . ($#credentials + 1)}) if ($debug >= 2);
+    push(@credentials, $args{"credential_" . ($#credentials + 1)});
+  }
+  
+  # Exchange the credentials for a session key.
+  my $ses_key = $self->authen_cred($r, @credentials);
+  $r->log_error("ses_key " . $ses_key) if ($debug >= 2);
+
+  # Send the Set-Cookie header.
+  my $cookie_path = $r->dir_config($auth_name . "Path");
+  $r->err_header_out("Set-Cookie" => "${auth_type}_${auth_name}=$ses_key; path=$cookie_path");
+
+  $r->no_cache(1);
+  $r->err_header_out("Pragma" => "no-cache");
+  $r->header_out("Location" => $args{'destination'});
+  return REDIRECT;
+}
+
+sub authenticate ($$) {
+  my ($auth_type, $r) = @_;
+  my ($authen_script, $auth_user);
+  my $debug = $r->dir_config("AuthCookieDebug") || 0;
+  
+  $r->log_error("auth_type " . $auth_type) if ($debug >= 3);
+  return OK unless $r->is_initial_req; # Only authenticate the first internal request
+  
+  if ($r->auth_type ne $auth_type) {
+    # This location requires authentication because we are being called,
+    # but we don't handle this AuthType.
+    $r->log_error($auth_type . "::Auth:authen auth type is " .
+		  $r->auth_type) if ($debug >= 3);
+    return DECLINED;
+  }
+
+  # Ok, the AuthType is $auth_type which we handle, what's the authentication
+  # realm's name?
+  my $auth_name = $r->auth_name;
+  $r->log_error("auth_name " . $auth_name) if $debug >= 2;
+  unless ($auth_name) {
+    $r->log_reason("AuthName not set, AuthType=$auth_type", $r->uri);
+    return SERVER_ERROR;
+  }
+
+  # There should also be a PerlSetVar directive that give us the path
+  # to set in Set-Cookie header for this realm.
+  my $cookie_path = $r->dir_config($auth_name . "Path");
+  unless ($cookie_path) {
+    $r->log_reason("Cookie path ($auth_name\Path) not set, AuthType=$auth_type", $r->uri);
+    return SERVER_ERROR;
+  }
+
+
+  # Get the Cookie header. If there is a session key for this realm, strip
+  # off everything but the value of the cookie.
+  my ($ses_key_cookie) = ($r->header_in("Cookie") || "") =~ /$auth_type\_$auth_name=([^;]+)/;
+  $ses_key_cookie = "" unless defined($ses_key_cookie);
+
+  $r->log_error("ses_key_cookie " . $ses_key_cookie) if ($debug >= 1);
+  $r->log_error("cookie_path " . $cookie_path) if ($debug >= 2);
+  $r->log_error("uri " . $r->uri) if ($debug >= 2);
+
+  if ($ses_key_cookie) {
+    if ($auth_user = $auth_type->authen_ses_key($r, $ses_key_cookie)) {
+      # We have a valid session key, so we return with an OK value.
+      # Tell the rest of Apache what the authentication method and
+      # user is.
+
+      $r->no_cache(1);
+      $r->err_header_out("Pragma", "no-cache");
+      $r->connection->auth_type($auth_type);
+      $r->connection->user($auth_user);
+      $r->log_error("user authenticated as $auth_user")	if $debug >= 1;
+      return OK;
+    } else {
+      # There was a session key set, but it's invalid for some reason. So,
+      # remove it from the client now so when the credential data is posted
+      # we act just like it's a new session starting.
+      
+      $r->err_header_out("Set-Cookie" => 
+			 "$auth_type\_$auth_name=; path=$cookie_path; expires=Mon, 21-May-1971 00:00:00 GMT");
+      $r->log_error("set_cookie " . $r->err_header_out("Set-Cookie"))
+	if $debug >= 2;
+    }
+  }
+
+  # They aren't authenticated, and they tried to get a protected
+  # document. Send them the authen form.  There should be a
+  # PerlSetVar directive that give us the name and location of the
+  # script to execute for the authen page.
+  
+  unless ($authen_script = $r->dir_config($auth_name . "LoginScript")) {
+    $r->log_reason($auth_type . 
+		   "::Auth:authen authentication script not set for auth realm " .
+		   $auth_name, $r->uri);
+    return SERVER_ERROR;
+  }
+  $r->custom_response(AUTH_REQUIRED, $authen_script);
+  
+  return AUTH_REQUIRED;
+}
+
+sub authorize ($$) {
+  my ($auth_type, $r) = @_;
+  my $debug = $r->dir_config("AuthCookieDebug") || 0;
+  
+  return OK unless $r->is_initial_req; #only the first internal request
+  
+  if ($r->auth_type ne $auth_type) {
+    $r->log_error($auth_type . "::Auth:authz auth type is " .
+		  $r->auth_type) if ($debug >= 3);
+    return DECLINED;
+  }
+  
+  my $reqs_arr = $r->requires or return DECLINED;
+  
+  my $user = $r->connection->user;
+  unless ($user) {
+    # user is either undef or =0 which means the authentication failed
+    $r->log_reason("No user authenticated", $r->uri);
+    return FORBIDDEN;
+  }
+  
+  my ($forbidden);
+  foreach my $req (@$reqs_arr) {
+    my ($requirement, $args) = split /\s+/, $req->{requirement}, 2;
+    $args = '' unless defined $args;
+    $r->log_error("requirement := $requirement, $args") if $debug >= 2;
+    
+    next if $requirement eq 'valid-user';
+    next if $requirement eq 'user' and $args =~ m/\b$user\b/;
+
+    # Call a custom method
+    my $ret_val = $auth_type->$requirement($r, $args);
+    $r->log_error("$auth_type->$requirement returned $ret_val") if $debug >= 3;
+    next if $ret_val == OK;
+
+    # Nothing succeeded, deny access to this user.
+    $forbidden = 1;
+    last;
+  }
+
+  return $forbidden ? FORBIDDEN : OK;
+}
+
+
 sub authen ($$) {
     my $that = shift;
     my $r = shift;
@@ -247,69 +424,103 @@ C<use mod_perl qw(1.07 StackedHandlers MethodHandlers Authen Authz);>
 <PRE>
 =end html
 
- # mod_perl startup script
-
+ # In mod_perl startup script:
  use Sample::AuthCookieHandler;
 
- # access.conf or .htaccess
+ # In httpd.conf or .htaccess:
 
- <Location /unprotected/protected>
-    PerlAuthenHandler Sample::AuthCookieHandler->authen
-    PerlAuthzHandler Sample::AuthCookieHandler->authz
-    AuthType Sample
-    AuthName WhatEver
-    PerlSetVar WhatEverPath /unprotected
-    PerlSetVar WhatEverLoginScript /unprotected/login.pl
-    require valid-user
+ PerlSetVar WhatEverPath /
+ PerlSetVar WhatEverLoginScript /login.pl
+
+ # These documents require user to be logged in.
+ <Location /protected>
+  AuthType Sample
+  AuthName WhatEver
+  PerlAuthenHandler Sample::AuthCookieHandler->authenticate
+  PerlAuthzHandler Sample::AuthCookieHandler->authorize
+  require valid-user
  </Location>
 
+ # These documents don't require logging in, but allow it.
+ <FilesMatch "\.ok$">
+  AuthType Sample
+  AuthName WhatEver
+  PerlFixupHandler Sample::AuthCookieHandler->recognize_user
+ </FilesMatch>
+
+ # This is the action of the login.pl script above.
+ <Files LOGIN>
+  AuthType Sample
+  AuthName WhatEver
+  SetHandler perl-script
+  PerlHandler Sample::AuthCookieHandler->login
+ </Files>
+
+
 =for html
 </PRE>
 =end html
 
 =head1 DESCRIPTION
 
-B<Apache::AuthCookie> allows you to intercept a users first
+B<Apache::AuthCookie> allows you to intercept a user's first
 unauthenticated access to a protected document. The user will be
-presented with a custom form where they can enter thier authentication
+presented with a custom form where they can enter authentication
 credentials. The credentials are posted to the server where AuthCookie
-verifies them and generates a session key.
+verifies them and returns a session key.
 
 The session key is returned to the user's browser as a cookie. As a
 cookie, the browser will pass the session key on every subsequent
 accesses. AuthCookie will verify the session key and re-authenticate
 the user.
 
-All you have to do is write a custom package that inheriets from
-AuthCookie.  Your package implements two functions:
+All you have to do is write a custom module that inherits from
+AuthCookie.  Your module implements two functions:
 
 =over 4
 
 =item C<authen_cred()>
 
-Verify the credentials and return a session key.
+Verify the user-supplied credentials and return a session key.  The
+session key can be any string - often you'll use some string
+containing username, timeout info, and any other information you need
+to determine access to documents, and append a one-way hash of those
+values together with some secret key.
 
 =item C<authen_ses_key()>
 
-Verify the session key and return the user ID.
+Verify the session key (previously generated by C<authen_cred()>,
+possibly during a previous request) and return the user ID.  This user
+ID will be fed to C<$r-E<gt>connection-E<gt>user()> to set Apache's
+idea of who's logged in.
 
 =back
 
-Using AuthCookie versus AuthBasic you get two key benefits.
+By using AuthCookie versus Apache's built-in AuthBasic you can design
+your own authentication system.  There are at least three main
+benefits.
 
 =over 4
 
 =item 1.
 
 The client doesn't *have* to pass the user credentials on every
-subsequent access. You have to do a little more work to get this
-feature, by having C<authen_cred()> generate a session key.
+subsequent access.  If you're using passwords, this means that the
+password can be sent on the first request only, and subsequent
+requests don't need to send this (potentially sensitive) information.
+This is known as "ticket-based" authentication.
 
 =item 2.
 
 When you determine that the client should stop using the
 credentials/session key, the server can tell the client to delete the
-cookie.
+cookie.  Letting users "log out" is a notoriously impossible-to-solve
+problem of AuthBasic.
+
+=item 3.
+
+AuthBasic dialog boxes are ugly.  You can design your own HTML login
+forms when you use AuthCookie.
 
 =back
 
@@ -338,15 +549,15 @@ client.
       /       \                                 V                   |
      /         \               +---------------------------------+  |
     /   Pass    \              | User's client submits this form |  |
-   /   user's    \             | to the same URL as the original |  |
-   | credentials |<------------| request. AuthCookie sees that   |  |
-   \     to      /             | the user isn't authenticated but|  |
-    \authen_cred/              | this time there are credentials |  |
-     \ function/               | as part of the request.         |  |
-      \       /                +---------------------------------+  |
+   /   user's    \             | to the LOGIN URL, which calls   |  |
+   | credentials |<------------| AuthCookie->login().            |  |
+   \     to      /             +---------------------------------+  |
+    \authen_cred/                                                   |
+     \ function/                                                    |
+      \       /                                                     |
        \     /                                                      |
         \   /            +------------------------------------+     |
-         \ /   return    | Authen cred returns an session     |  +--+
+         \ /   return    | Authen cred returns a session      |  +--+
           V------------->| key which is opaque to AuthCookie.*|  |
                 True     +------------------------------------+  |
                                               |                  |
@@ -364,7 +575,7 @@ client.
                |                    +-| authen_ses_key|------------+
                V                       \             /     False
   +-----------------------------------+ \           /
-  | Tell Apache to set Expires header |  \         /
+  | Tell Apache to set Expires header,|  \         /
   | set no-cache Pragma header, set   |   \       /
   | user to user ID returned by       |    \     /
   | authen_ses_key, set authentication|     \   /
@@ -373,27 +584,48 @@ client.
          (---------------------)              ^
          ( Request a protected )              |
          ( page, user has a    )--------------+
-         ( session key cookie  )                    
+         ( session key cookie  )
          (---------------------)
 
 
- *  The session key that the client gets can be anything you
-    want. For example, encrypted information about the user, a
-    hash of the username and password (similar in function to
-    Digest authentication), or the user name and password in
-    plain text (similar in function to HTTP Basic
-    authentication).
+ *  The session key that the client gets can be anything you want.  For
+    example, encrypted information about the user, a hash of the
+    username and password (similar in function to Digest
+    authentication), or the user name and password in plain text
+    (similar in function to HTTP Basic authentication).
 
-    They only catch is that the authen_ses_key function that you
-    create must be able to determine if this session_key is valid
-    and map it back to the originally authenticated user ID.
+    The only requirement is that the authen_ses_key function that you
+    create must be able to determine if this session_key is valid and
+    map it back to the originally authenticated user ID.
 
 =for html
 </PRE>
 
+=head1 UPGRADING FROM VERSION 1.4
+
+There are a couple of interface changes that you need to be aware of
+when migrating from version 1.4 to 2.0.  First, the authen() and
+authz() methods are now deprecated, replaced by the new authenticate()
+and authorize() methods.  The old methods will go away in a couple
+versions, but are maintained intact in this version to ease the task
+of upgrading.  The use of these methods is essentially the same, though.
+
+Second, when you change to the new method names (see previous
+paragraph), you must change the action of your login forms to the
+location /LOGIN (or whatever URL will call your module's login()
+method).
+
+Third, you must change your login forms (see L<THE LOGIN SCRIPT>
+below) to indicate how requests should be redirected after a
+successful login.
+
+
 =head1 EXAMPLE
 
-Try the sample in eg/.
+NOTE: THE FOLLOWING EXAMPLE CODE IS BASED ON AN OLD VERSION OF
+Apache::AuthCookie, AND IS NO LONGER SUPPORTED.  I'll update the
+example stuff as soon as I get a chance.  For now, please see the
+sample configuration code at the top of this document.
 
 =head2 Install the sample
 
@@ -420,7 +652,7 @@ Restart Apache so mod_perl picks up C<Sample::AuthCookieHandler>.
 
 =back
 
-=head2 Tryout the sample
+=head2 Try out the sample
 
 =over 4
 
@@ -441,7 +673,35 @@ line C<require valid-user> to C<require dwarf> in .htaccess only the
 user "programmer" will have access. Look at the authorization function
 C<dwarf()> in F<eg/Sample/AuthCookieHandler.pm> to see how this works.
 
-=head2 Using a Session Key
+=back
+
+=head1 THE LOGIN SCRIPT
+
+You will need to create a login script (called login.pl above) that
+generates an HTML form for the user to fill out.  The following fields
+must be present in the form:
+
+=over 4
+
+=item 1.
+
+The ACTION of the form must be /LOGIN (or whatever you defined in your
+server configuration, as in the SYNOPSIS section).
+
+=item 2.
+
+The various user input fields (username, passwords, etc.) must be
+named 'credential_0', 'credential_1', etc. on the form.
+
+=item 3.
+
+You must define a form field called 'destination' that tells
+AuthCookie where to redirect the request after successfully logging
+in.  Typically this value is obtained from C<$r-E<gt>prev-E<gt>uri>.
+
+=back
+
+=head1 ABOUT SESSION KEYS
 
 Unlike the sample AuthCookieHandler, you have you verify the user's
 login and password in C<authen_cred()>, then you do something
@@ -450,7 +710,7 @@ like:
     my $date = localtime;
     my $ses_key = MD5->hexhash(join(';', $date, $PID, $PAC));
 
-save C<$ses_key> alogin with the user's login, and return C<$ses_key>.
+save C<$ses_key> along with the user's login, and return C<$ses_key>.
 
 Now C<authen_ses_key()> looks up the C<$ses_key> passed to it and
 returns the saved login.  I use Oracle to store the session key and
@@ -458,8 +718,12 @@ retrieve it later, see the ToDo section below for some other ideas.
 
 =head1 KNOWN LIMITATIONS
 
-The first unauthenticated request can not be a POST. B<Apache::AuthCookie>
-interupts the initial request with the form for the user credentials.
+The login form uses the GET method to send the user's credentials, so
+they're visible on the query string.
+
+If the first unauthenticated request is a POST, it will be changed to
+a GET after the user fills out the login forms, and POSTed data will
+be lost.
 
 =head2 ToDo
 
@@ -467,24 +731,23 @@ interupts the initial request with the form for the user credentials.
 
 =item *
 
-See if there's a way to allow the initial request to be a POST.  I know
-B<Apache::AuthCookie> could save the original POST'ed material but I
-don't know how to re-insert this content into the request stream after
+There ought to be a way to solve the GET/POST problems in the
+LIMITATIONS section.  They both involve being able to change a request
+back & forth between GET & POST.  The second problem also involves
+being able to re-insert the POSTed content into the request stream after
 the user authenticates.  If you knows of a way, please drop me a note.
 
-=item *
-
-Create a session key store that uses shared memory.  If anyone wants to
-get together and help with this that would be cool.  Storing and
-retrieving them should be easy, but the harder part is cleaning out
-"old" keys and dealing with server restarts without losing all the
-keys.
+In the next release there will be a logout() method that will greatly
+simplify the process of deleting someone's cookie.  I wanted to get
+this release out the door though, so I didn't put it in this one.
 
 =back
 
 =head1 AUTHOR
 
-Eric Bartley, bartley@purdue.edu
+Ken Williams, ken@forum.swarthmore.edu
+
+Originally written by Eric Bartley, bartley@purdue.edu
 
 =head1 SEE ALSO
 
